unit SPCXImg;

{ Freeware PCX image  Delphi-component for both 16 and 32 bit
  made by Endre I. Simay, Hungary, 1998.

  It's made by rewriting the LoadFromStream procedure.
  There was using some snipets cut from
  SHOW_PCX Source [(C) Copyright 1989 ZSoft, Corporation.].
  during evaluation of TSPCX256 (descendant of TPersistent)

  This is an enhanced component as a descendant from
  TUniBitmap (made by same author) and register the PCX file format, 
  so You Can load a PCX from a Resource (*.RES) file
  as RCDATA.

  The supported images should be have 16, 256 or 24bit colors.

Left Open :
  - Also store the format via SaveToStream override
    so the PCX format can be used in blob formats
    also !
  - Older CGA PCXs format not work propably
  - Slow at large 24bit color images
}

interface
Uses
Forms, WinTypes, WinProcs, Graphics, SysUtils, Classes,
UniBmp;

const ErrName: Array[1..3] of string = (
        'This is not a valid PCX image file.',
        'This is not a 16, 256 or 24bit color PCX image file.',
        'PCX file is in wrong format - It must be an'+{ CGA,}' EGA, VGA, MCGA or 24bit image.'
        );

CONST
 MAX_WIDTH = 2*4000;    { arbitrary - maximum width (in bytes) of a PCX image }
 COMPRESS_NUM = $C0;  { this is the upper two bits that indicate a count }
 MAX_BLOCK = 2*4096;

 RED = 0;
 GREEN = 1;
 BLUE = 2;

 CGA4 = $04;          { video modes }
 CGA2 = $06;
 EGA  = $10;
 VGA  = $12;
 MCGA = $13;
 MCGA2 = $15;

type
{ Integer16 will be 16bit both on 16 and 32 bit Windows' }
{$IFDEF WIN32}
Integer16=Smallint;
{$ELSE}
Integer16=Integer;
{$ENDIF}
type
 block_ARRAY = ARRAY [0..MAX_BLOCK] OF Byte;
 pal_ARRAY = ARRAY [0..255, RED..BLUE] OF Byte;
 line_ARRAY =packed ARRAY [0..MAX_WIDTH] OF Byte;

type
 PBmLine         = ^TBmpLinesStruct;
 TBmpLinesStruct = packed record
  LineData  : Line_ARRAY;
  LineNo    : LongInt;
  end;

 TPCX_Header = Packed RECORD
  Manufacturer: Byte;     { Always 10 for PCX file }
  Version: Byte;          { 2 - old PCX - no palette (NOT used anymore),
                            3 - no palette,
                            4 - Microsoft Windows - no palette (only in
                                old files, New Windows version USES 3),
                            5 - WITH palette }
  Encoding: Byte;         { 1 is PCX, it is possible that we may add
                              additional encoding methods IN the future }
  Bits_per_pixel: Byte;   { Number of bits to represent a pixel
                            (per plane) - 1, 2, 4, or 8 }
  Xmin: Integer16;          { Image window dimensions (inclusive) }
  Ymin: Integer16;          { Xmin, Ymin are usually zero (not always) }
  Xmax: Integer16;
  Ymax: Integer16;
  Hdpi: Integer16;          { Resolution of image (dots per inch) }
  Vdpi: Integer16;          { Set to scanner resolution - 300 is default }
  ColorMap: ARRAY [0..15, RED..BLUE] OF Byte;
                          { RGB palette data (16 colors or less)
                            256 color palette is appended to END OF FILE }
  Reserved: Byte;         { (used to contain video mode)
                            now it is ignored - just set to zero }
  Nplanes: Byte;          { Number of planes }
  Bytes_per_line_per_plane: Word;
                          { Number of bytes to allocate
                            for a scanline plane.
                            MUST be an an EVEN number!
                            DO NOT calculate from Xmax-Xmin! }
  PaletteInfo: Word;      { 1 = black & white or color image,
                            2 = grayscale image - ignored IN PB4, PB4+
                            palette must also be set to shades of gray! }
  HscreenSize: Integer16;   { added for PC Paintbrush IV Plus ver 1.0,  }
  VscreenSize: Integer16;   { PC Paintbrush IV ver 1.02 (and later)     }
                          { I know it is tempting to use these fields
                            to determine what video mode should be used
                            to display the image - but it is not
                            recommended since the fields will probably
                            just contain garbage. It is better to have
                            the user install for the graphics mode he
                            wants to use... }
 Filler: ARRAY [74..127] OF Byte;     { Just set to zeros }
 END;

{VAR}

type

 EPCXException = class(Exception)
  end;

 TSPCX256 = Class (TPersistent)
  private
   BlockData: block_ARRAY;     { 8k data buffer }
   Palette256: pal_ARRAY;      { place to put 256 color palette }
   PCXline: line_ARRAY;        { place to put uncompressed data }
   Ymax: Integer16;            { maximum Y value on screen }
   NextByte: Integer16;        { index into file buffer in ReadByte }
   Index: Integer16;           { PCXline index - where to put Data }
   Data: Byte;                 { PCX compressed data byte }
   PictureMode: Integer16;     { Graphics mode number }
   FStream    : TStream;            { the file stream for the pcx file }
   FOStream   : TMemoryStream;      { the file output stream for the image }
   Header     : TPCX_Header;        { pcx file header }
   BmHeader   : TBitmapInfoHeader;  { File Header for bitmap file}
   BmFHeader  : TBitmapFileHeader;  { File Header for bitmap file}
   ImageLines : TList;              { Image data}
   {LineBuffer : line_ARRAY;         { array for buffer line output }
   CurrentY   : LongInt;            { current screen-line location }
  protected
  public
   constructor Create;
   destructor  Destroy; override;
   procedure   Error(ErrCode: Integer16);
   procedure   EntireVGApalette;
   procedure   SetPalette;
   procedure   CreateLine;
   procedure   DecodeVGA;
   procedure   DecodeMCGA;
   procedure   DecodeMCGA2;
   procedure   VGA16palette;
   procedure   Read256palette;
   procedure   ReadByte;
   procedure   Read_PCX_Line;
   procedure   Read_PCX;
   procedure   ReadHeader;
   procedure   LoadFromStream(Stream: TStream);
   procedure   SaveToStream(Stream: TStream);
  published
  end;

 TSPCXBitmap = Class (TUniBitmap)
  public
   procedure LoadFromStream(Stream: TStream); override;
  end;

implementation

procedure TSPCXBitmap.LoadFromStream(Stream: TStream);
var
  DBmp:TBitmap;
  SP:TSPCX256;
  MS:TMemoryStream;
begin
  DBmp:=TBitmap.Create;
  SP:=TSPCX256.Create;
  MS:=TMemoryStream.Create;
  try
   SP.LoadFromStream(Stream);
   SP.SaveToStream(MS);
   MS.Seek(0,0);
   DBmp.LoadFromStream(MS);
   Assign(DBmp);
  finally
   DBmp.Free;
   SP.Free;
   MS.Free;
  end;
end;

procedure TSPCX256.Error(ErrCode: Integer16);
begin
  raise EPCXException.Create(ErrName[ErrCode]);
end;

constructor TSPCX256.Create;
begin
 Inherited Create;
 FOStream := TMemoryStream.Create;
 FStream := nil;
 ImageLines := TList.Create;
end;

destructor  TSPCX256.Destroy;
begin
 FOStream.Free;
 ImageLines.Free;
 Inherited Destroy;
end;

procedure TSPCX256.LoadFromStream(Stream: TStream);
begin
  FStream:=Stream;
  Read_PCX; { read and display the file }
  FOStream.Seek(0,0);
end;

procedure TSPCX256.SaveToStream(Stream: TStream);
begin
FOStream.Seek(0,0);
TMemoryStream(Stream).LoadFromStream(FOStream);
end;

procedure TSPCX256.Read256palette;
{ Read in a 256 color palette at end of PCX file }
VAR
  b: Byte;
BEGIN
 FStream.Seek(FStream.Size - 769,0);
 FStream.Read (b, 1);                   { read indicator byte }
 IF b <> 12 THEN                        { no palette here... }
   Exit;
 FStream.Read(Palette256, 3*256);
 FStream.Seek(128,0);                 { go back to start of PCX data }
END;  { Read256palette }

procedure TSPCX256.ReadHeader;
LABEL Wrongformat;
begin
  FStream.Seek (0,0);
  FStream.Read (Header, 128); { read 128 byte PCX header }
  IF (Header.Manufacturer <> 10) OR (Header.Encoding <> 1) THEN
    BEGIN
     Error(1);
    END;
  IF (Header.Nplanes = 3) AND (Header.Bits_per_pixel = 8) THEN
             PictureMode := MCGA2
             else
  IF (Header.Nplanes = 4) AND (Header.Bits_per_pixel = 1) THEN
    BEGIN
     PictureMode :=VGA;
    END
     ELSE IF (Header.Nplanes = 1) THEN
      BEGIN
      IF (Header.Bits_per_pixel = 1) THEN
       Error(2)
       ELSE IF (Header.Bits_per_pixel = 2) THEN
           Error(2)
           ELSE IF (Header.Bits_per_pixel = 8) THEN
            begin
             PictureMode := MCGA;
              IF Header.Version = 5 then
                 Read256palette;
            end
            ELSE
             GOTO Wrongformat;
        END
        ELSE
        BEGIN
          Wrongformat:
          Error(3);
        END;
   Index := 0;
   NextByte := MAX_BLOCK;          { indicates no data read in yet... }
end;

PROCEDURE TSPCX256.VGA16palette;
VAR
       i: INTEGER;
       l: LongInt;
const  j: Byte=0;
BEGIN
  with BmHeader do
     begin
      biBitCount       := 4;
     end;

  with BmFHeader do begin
     L:=(Header.Ymax-Header.Ymin+1);
     L:=L*(Header.Xmax-Header.Xmin+1)div 2 ;
     bfSize :=BmHeader. biSizeImage+
     longint((3*15) +
     Sizeof(TBitmapFileHeader) +  {Color map info}
     Sizeof(TBitmapInfoHeader) );

     bfReserved1 := 0; {not currently used}
     bfReserved2 := 0; {not currently used}
     bfOffBits := (4*16)+ Sizeof(TBitmapFileHeader)+
                           Sizeof(TBitmapInfoHeader);
   end;
   FOStream.Seek(0,0);
   FOStream.Write(BmFHeader,SizeOf(TBitmapFileHeader));
   FOStream.Write(BmHeader,SizeOf(TBitmapInfoHeader));
   FOR i := 0 TO 15 DO
     BEGIN                                          { R, G, and B must be 0..63 }
      FOStream.Write( Header.ColorMap  [i, BLUE],1);
      FOStream.Write( Header.ColorMap  [i, GREEN],1);
      FOStream.Write( Header.ColorMap  [i, RED],1);
      FOStream.Write( j,1);
     end;
end;

procedure TSPCX256.EntireVGApalette;
VAR
 i: INTEGER;
const j:Byte=0;
BEGIN
  with BmHeader do
     begin
      biBitCount := 8;
     end;

  with BmFHeader do
   begin
     bfSize :=
     longint((3*255) + Sizeof(TBitmapFileHeader) +  {Color map info}
                         Sizeof(TBitmapInfoHeader) +
        ((Header.Ymax-Header.Ymin+1)*(Header.Xmax-Header.Xmin+1)));
     bfReserved1 := 0; {not currently used}
     bfReserved2 := 0; {not currently used}
     bfOffBits := (4*256)+ Sizeof(TBitmapFileHeader)+
                           Sizeof(TBitmapInfoHeader);
   end;
   FOStream.Seek(0,0);
   FOStream.Write(BmFHeader,SizeOf(TBitmapFileHeader));
   FOStream.Write(BmHeader,SizeOf(TBitmapInfoHeader));
   for i := 0 TO 255 DO
    BEGIN                                          { R, G, and B must be 0..63 }
     FOStream.Write( Palette256 [i, BLUE],1);
     FOStream.Write( Palette256 [i, GREEN],1);
     FOStream.Write( Palette256 [i, RED],1);
     FOStream.Write( j,1);
    END;
end;

procedure TSPCX256.SetPalette;
{ Set up the entire graphics palette }
BEGIN
 IF PictureMode = MCGA THEN
   EntireVGApalette
 ELSE
 IF PictureMode = VGA THEN
   VGA16palette
 ELSE
 IF PictureMode <> MCGA2 THEN
 Error(2);
END;  { SetPalette }

procedure TSPCX256.ReadByte;
{ read a single byte of data - use BlockRead because it is FAST! }
VAR
 NumBlocksRead: Word;
BEGIN
 IF NextByte = MAX_BLOCK THEN
  BEGIN
   NumBlocksRead:=FStream.Read ( BlockData, MAX_BLOCK);
   NextByte := 0;
  END;
   data := BlockData [NextByte];
   Inc (NextByte);   { NextByte++; }
END;  { ReadByte }

procedure TSPCX256.Read_PCX_Line;
{ Read a line from a PC Paintbrush PCX file }
VAR
 count: Word;
 bytes_per_line: Word;
BEGIN
{$I-}
  bytes_per_line :=( Header.Bytes_per_line_per_plane * Header.Nplanes);
  { bring in any data that wrapped from previous line }
  { usually none  -  this is just to be safe          }
  IF Index <> 0 THEN
    FillChar (PCXline [0], Index, data);     { fills a contiguous block of data }
  WHILE (Index < bytes_per_line) DO          { read 1 line of data (all planes) }
    BEGIN
     ReadByte;
     IF (data AND $C0) = compress_num THEN
       begin
        count := data AND $3F;
        ReadByte;
        FillChar (PCXline [Index], count, data);  { fills a contiguous block }
        Inc (Index, count);                       { Index += count; }
       end
      ELSE
       begin
        PCXline [Index] := data;
        Inc (Index);                              { Index++; }
       END;
    END;
    Index := Index - bytes_per_line;
END;  { Read_PCX_Line }

procedure TSPCX256.CreateLine;
var
  p: PBmLine;
begin
  Application.ProcessMessages;
  {Create a new bmp line}
  New(p);
  p^.LineData:= PCXLine;
  p^.LineNo := CurrentY;
  ImageLines.Add(p);
  {Prepare for the next line}
  Inc(CurrentY);
end;

PROCEDURE TSPCX256.DecodeVGA;
VAR
  k, kmax: Integer16;
  j,i,y: LongInt;
  c:byte;
  Line: LongInt;
  ch: char;
  p: ^TBmpLinesStruct;
  x: LongInt;
  kk,Ly,Maxx,Lx:Integer;
  la:Line_array;
begin
   kmax := Header.Ymin + Header.ymax;
   FOR k := Header.Ymin TO kmax DO
     begin
      Read_PCX_Line;
      FillChar(la [0],sizeof(la),0);
      x:=0;
      FOR j:=0 TO Header.Nplanes-1 DO
        begin
        FOR i:=0 TO Header.Bytes_per_line_per_plane-1 DO
          BEGIN
           c:=pcxline[x];
           inc(x);
           FOR kk:=0 TO 7 DO
             IF (c AND (128 SHR kk))>0
           THEN la[(i*8)+kk]:=la[(i*8)+kk] OR 1 SHL j;
          END;
        end;
      I:=-1;
      x:=-1;
      While x<=(Header.Xmax - Header.Xmin) DO
       begin
        Inc(i);
        Inc(x);
        pcxline[i]:=(la[x] SHL 4)XOR la[x+1] AND $0F;
        x:=x+1;
       end;
      CreateLine;
     END;

  Maxx:=((Header.Xmax - Header.Xmin)+1)div 2;
  Lx:=Maxx DIV 4; Ly:=4-(Maxx-Lx*4);
  if (ly=4)  then
   begin
    ly:=0;
   end;
  Kk:=Maxx+Ly;
  FOR k := (ImageLines.Count - 1) downto 0 DO
   begin
     p := ImageLines.Items[k];
     FOStream.Write(p^.LineData,kk);
  end;
end;

procedure TSPCX256.DecodeMCGA;
VAR
  k, kmax: Integer16;
  p: ^TBmpLinesStruct;
  kk,Ly,Maxx,Lx:Integer;
begin
  kmax := Header.Ymin +Header.ymax;
  for k := Header.Ymin TO kmax DO
   begin
    Read_PCX_Line;
    CreateLine;
   END;

  Maxx:=(Header.Xmax - Header.Xmin)+1;
  Lx:=Maxx DIV 4; Ly:=4-(Maxx-Lx*4);
  Ly:=Ly*Byte(Ly<>4);

  Kk:=Maxx+Ly;

  for k := (ImageLines.Count - 1) downto 0 DO
   begin
     p := ImageLines.Items[k];
     FOStream.Write(p^.LineData,kk);
   end;
end;

procedure TSPCX256.DecodeMCGA2;
VAR
  k, kmax: Integer16;
  p: ^TBmpLinesStruct;
  x: LongInt;
  kk,Ly,Maxx,Lx:Integer;
 L: LongInt;
 La:Line_Array;
const  j:Byte=0;
BEGIN
  with BmHeader do
     begin
      biBitCount := 24;
     end;

  with BmFHeader do
   begin
     L:=(Header.Ymax-Header.Ymin+1);
     L:=L*(Header.Xmax-Header.Xmin+1)*3;
     bfSize :=L;
     bfSize :=bfSize+
      Sizeof(TBitmapFileHeader) +  {Color map info}
                         Sizeof(TBitmapInfoHeader);
     bfReserved1 := 0; {not currently used}
     bfReserved2 := 0; {not currently used}
     bfOffBits := Sizeof(TBitmapFileHeader)+
                           Sizeof(TBitmapInfoHeader);
   end;
   FOStream.Seek(0,0);
   FOStream.Write(BmFHeader,SizeOf(TBitmapFileHeader));
   FOStream.Write(BmHeader,SizeOf(TBitmapInfoHeader));

  kmax := Header.Ymin +Header.ymax;
  for k := Header.Ymin TO kmax DO
   begin
    Read_PCX_Line;
    x:=0;
    for L:=0 to Header.Bytes_per_line_per_plane do
     begin
      la[x+2]:=PCXLine[l];
      la[x+1]:=PCXLine[l+{Maxx}Header.Bytes_per_line_per_plane];
      la[x]:=PCXLine[l+2*{Maxx}Header.Bytes_per_line_per_plane];
      la[x+3]:=0;
      x:=x+3;
     end;
    pcxline:=la;
    CreateLine;
   END;

  Maxx:=3*((Header.Xmax - Header.Xmin)+1);
  Lx:=Maxx DIV 4; Ly:=4-(Maxx-Lx*4);
  Ly:=Ly*Byte(Ly<>4);

  Kk:=Maxx+Ly;

  for k := (ImageLines.Count - 1) downto 0 DO
   begin
     p := ImageLines.Items[k];
     FOStream.Write(p^.LineData,kk);
   end;
end;

procedure TSPCX256.Read_PCX;
begin
  CurrentY :=0;                                            { current screen locations }
  ReadHeader;     { read the PCX header }
  ImageLines.Clear;
  ImageLines.Capacity :=Header.YMax-Header.YMin+1;
  with BmHeader do
   begin
    biSize           := Sizeof(TBitmapInfoHeader);
    biWidth          := (Header.Xmax-Header.Xmin+1);
    biHeight         := (Header.Ymax-Header.Ymin+1);
    biPlanes         := 1;            {Arcane and rarely used}
    biBitCount       := 8;
    biCompression    := BI_RGB;       {Sorry Did not implement compression in this version}
    biSizeImage      := 0;            {Valid since we are not compressing the image}
    biXPelsPerMeter  := 143;          {Rarely used very arcane field}
    biYPelsPerMeter  := 143;          {Ditto}
    biClrUsed        := 0;            {all colors are used}
    biClrImportant   := 0;            {all colors are important}
   end;
  BmFHeader.bfType:=
   (ORD('M') SHL 8)+ORD('B');

  IF Header.Version = 5 THEN
   SetPalette;                         { set the screen palette, if available }
   IF (PictureMode = VGA) THEN
     BEGIN
      DecodeVGA;
     END
    ELSE
   IF (PictureMode = MCGA) THEN
     BEGIN
      DecodeMCGA;
     END
    ELSE
    if (PictureMode = MCGA2) then
     DecodeMCGA2
     else
      Error(2); {e.g. CGA}                                      { it's a CGA picture }
end;

initialization
  { register the TSPCXBitmap as a new graphic file format
    for 16, 256 and 24bit color PCX-images for read but not write it;
    now all the TPicture storage stuff can access our new
    PCX graphic format ! }
  TPicture.RegisterFileFormat('pcx','PCX-Format', TSPCXBitmap);
end.
